//pin 1 & 2 as used for RX/TX
const int DATA0_PIN = 32;
const int DATA1_PIN = 33; 
const int DATA2_PIN = 25; 
const int DATA3_PIN = 26; 
const int DATA4_PIN = 27; 
const int DATA5_PIN = 14; 
const int DATA6_PIN = 12; 
const int DATA7_PIN = 13; 

const int DEBUG_PIN = 2;
const int DEBUG2_PIN = 15;
const int IR_PIN = 19;
const int RD_PIN = 5;
const int WR_PIN = 0; 


// logic analyzer library for DMA reading of 8-bit bus
#include "ESP32_LogicAnalyzer.h"
i2s_parallel_buffer_desc_t bufdesc;
i2s_parallel_config_t cfg;


// Tektronix 150xy RS232-plugin
const byte DEVICE_ID= 0x81;
int sequenceStep = 0;
int startStep = 0;
int previousStep = 0;
unsigned long stepTime;

bool acceptFrameDirective = false;
bool sendFrameDirective = false;
bool resetDirective = false; 

// set interrupts for bus
bool inputReady =false;
byte inputPortB= 0;
byte inputPortD=0;
bool outputReady =false;
byte inputByte;
byte outputByte;

// Reading
const byte NUMRECEIVE = 550;
char receivedBytes[NUMRECEIVE];
int receiveCounter = 0;
long gpio_reg0,gpio_reg1; 

// Writing
const byte NUMSEND = 550;
char sendBytes[NUMSEND];
int sendCounter = 0;
int sendLength = 0;

void IRAM_ATTR ProcessInput() //must be declared before setup
{
  //gpio_reg0 = REG_READ(GPIO_IN_REG); // GPIO 0..31
  gpio_reg0 = GPIO.in;
  //gpio_reg1 = REG_READ(GPIO_IN1_REG); // GPIO 32..39
  
  inputByte=0;

  if (bitRead(gpio_reg0,25))   digitalWrite(DEBUG2_PIN,HIGH); 
  else   digitalWrite(DEBUG2_PIN,LOW); 

  digitalWrite(DEBUG_PIN, !digitalRead(DEBUG_PIN)); 
  
//
//
//  if( digitalRead(DATA0_PIN)) bitSet(inputByte,0);
//  if( digitalRead(DATA1_PIN)) bitSet(inputByte,1);
//  if( digitalRead(DATA2_PIN)) bitSet(inputByte,2);
//  if( digitalRead(DATA3_PIN)) bitSet(inputByte,3);
//  if( digitalRead(DATA4_PIN)) bitSet(inputByte,4);
//  if( digitalRead(DATA5_PIN)) bitSet(inputByte,5);
//  if( digitalRead(DATA6_PIN)) bitSet(inputByte,6);
//  if( digitalRead(DATA7_PIN)) bitSet(inputByte,7); 
//  inputReady =true;
}

void setup() {
  
   //Serial config
   Serial.begin(19200);
   
   // set inputs & outputs
   readMode();    
   pinMode(RD_PIN, INPUT_PULLUP); 
   pinMode(WR_PIN, INPUT_PULLUP);
      
   pinMode(DEBUG_PIN, OUTPUT); 
   pinMode(DEBUG2_PIN, OUTPUT);     
   pinMode(IR_PIN, OUTPUT);

   attachInterrupt(digitalPinToInterrupt(WR_PIN),  ProcessInput, FALLING );
   attachInterrupt(digitalPinToInterrupt(RD_PIN),  ProcessOutput, FALLING );

   //no service request yet
   digitalWrite(IR_PIN,HIGH); 
   digitalWrite(DEBUG_PIN,LOW);     
   digitalWrite(DEBUG2_PIN,LOW);     

   //logic analyzer config 
    pinMode(ledPin, OUTPUT);
    
    dma_desc_init(CAPTURE_SIZE);
  
    cfg.gpio_bus[0] = 0;
    cfg.gpio_bus[1] = 18;//GPIO01 used for UART 0 RX
    cfg.gpio_bus[2] = 2;
    cfg.gpio_bus[3] = 19;//GPIO03 used for UART 0 TX
    cfg.gpio_bus[4] = 4;
    cfg.gpio_bus[5] = 5;
    cfg.gpio_bus[6] = -1;//GPIO06 used for SCK, bootloop
    cfg.gpio_bus[7] = -1;//GPIO07 used for SDO, bootloop
  
    cfg.gpio_bus[8] = -1;//GPIO8 used for SDI, bootloop
    cfg.gpio_bus[9] = -1;//GPIO9 lead SW_CPU_RESET on WROOVER module
    cfg.gpio_bus[10] = -1;//GPI10 lead SW_CPU_RESET on WROOVER module
    cfg.gpio_bus[11] = 22;//GPIO11 used for CMD, bootloop
    cfg.gpio_bus[12] = 12;
    cfg.gpio_bus[13] = 13;
    cfg.gpio_bus[14] = 14;
    cfg.gpio_bus[15] = 15;
  
  
    cfg.gpio_clk = 23; // Pin23 used for XCK input from LedC
    cfg.bits = I2S_PARALLEL_BITS_16;
    cfg.clkspeed_hz = 2 * 1000 * 1000; //resulting pixel clock = 1MHz
    cfg.buf = &bufdesc;

  //enable_out_clock(I2S_HZ);
  //fill_dma_desc( bufdesc );
  i2s_parallel_setup(&cfg);
  
//   Serial.println("Start up finished");
//   Serial.println(getCpuFrequencyMhz());
}

void loop() {
  
      //Serial.println(gpio_reg0);
      digitalWrite(DEBUG2_PIN,LOW);

      //CheckSerial();
      startStep = sequenceStep; 

      //Serial.println(sequenceStep);       
      switch (sequenceStep) {
     
      //********************
      // POLL Service
      //********************          
      case 0: // wait for RD and WR to be high before requesting service
        UpdateStepTime();
        digitalWrite(IR_PIN,HIGH); // reset service request        
        if(micros()>stepTime+300) sequenceStep = 10;           
        break;
      
      case 10: // service request by device 
        UpdateStepTime();
        outputReady =false;   
        writeByte(DEVICE_ID); // set deviceID on data bus      
        digitalWrite(IR_PIN,LOW); // set service request
 
        sequenceStep = 20;            
        break;
       
      case 20: // wait for read device-ID
        UpdateStepTime();
        CheckStepTime(500000); // indifinite     
        if(outputReady) // device-ID is read
        {
          outputReady = false;
          inputReady = false;
          readMode();
          sequenceStep = 30;
        }
        break;

      case 30: // wait for directive
      
        UpdateStepTime();
        CheckStepTime(75); // 1 < x < 50µs
        
        if(inputReady) // directive is written
        {
          inputReady=false;
          readDirective(inputByte);
          if(acceptFrameDirective) 
          {
            sequenceStep = 100;  
            //digitalWrite(DEBUG_PIN, !digitalRead(DEBUG_PIN));  
          }
          else if (sendFrameDirective) 
          {
            sequenceStep = 200;
            //digitalWrite(DEBUG2_PIN, !digitalRead(DEBUG2_PIN));
          }
          else if (resetDirective)  ResetSequence();                
        }
        break;
        
      //***********************
      // ACCEPT Frame Sequence
      //***********************
      case 100: // start receiving bytes
        UpdateStepTime();
        receiveCounter = 0;
        
        digitalWrite(IR_PIN,HIGH); // release service request
     
        sequenceStep = 110;
        break;

      case 110: // wait for byte-n on bus or end of transmission
        UpdateStepTime();
       
        if(micros()>stepTime+300) sequenceStep = 130; // end of transmission
        if(inputReady) // byte is ready to be read
        { 
          inputReady = false;
          receivedBytes[receiveCounter] = inputByte;
          receiveCounter+=1;
          sequenceStep = 120;
        }
        break;

      case 120: // prepare for next btype
        UpdateStepTime();
        sequenceStep = 110;
        break;

      case 130: //end of read Frame
        UpdateStepTime();
        Serial.write(receivedBytes, receiveCounter);
        sequenceStep = 0;
        break;

      //**********************
      // WRITE Frame Sequence
      //**********************
      case 200: // start writing bytes or end transmission immediatly
        UpdateStepTime();
        digitalWrite(IR_PIN,LOW); // do not release service request
        
        if(sendLength>0) 
        {      
          sendCounter = 0;
          sequenceStep = 210;
          outputReady= false;
        }
        else  sequenceStep = 350; // end of transmission - no data
        break;
    
       case 210: // write byte to bus, wait for it to be read
        UpdateStepTime();
        CheckStepTime(250); // 150 < x < 200 µs
        writeByte(sendBytes[sendCounter]);          
        if(outputReady)  
        {    
          outputReady= false;
          sequenceStep = 220;   
        }
  
        break;

      case 220: //prepare for next byte
        UpdateStepTime();
        CheckStepTime(75); // 1 < x < 50µs
        
        sendCounter+=1;
        if(sendCounter >=sendLength)sequenceStep = 250; // end of transmission
        else  
        {
          outputReady =false;
          sequenceStep = 210; // next byte   
        }
        
        break;

      case 250: // end of transmission 
        UpdateStepTime();
        
        sendLength = 0; 
        digitalWrite(IR_PIN,HIGH); // service request must be released upon receiving the last byte in the frame
        readMode();

        if(micros()>stepTime+300) sequenceStep = 0;
        break;

       case 350: // nothing to write yet
        UpdateStepTime();
        
        digitalWrite(IR_PIN,HIGH);   
        readMode();

        if(micros()>stepTime+300) sequenceStep = 360;
        break;
   
      default: // something went wrong
        ResetSequence();
      }  
      previousStep = startStep;
}
 
void ResetSequence()
{
   sequenceStep = 0;
   receiveCounter = 0;
   sendCounter = 0;
   sendLength = 0;
   readMode();
   digitalWrite(IR_PIN,HIGH);
   delay(20);//at least 20ms should be wait
}

void ResetSequenceTimeOut()
{
   sequenceStep = 0;
   receiveCounter = 0;
   if(sendCounter != 0) 
   {
     sendCounter = 0;
     sendLength = 0;
   }
   readMode();
   digitalWrite(IR_PIN,HIGH);
}

void writeByte(byte b)
{
  writeMode();
  digitalWrite(DATA0_PIN,bitRead(b,0));
  digitalWrite(DATA1_PIN,bitRead(b,1));
  digitalWrite(DATA2_PIN,bitRead(b,2));
  digitalWrite(DATA3_PIN,bitRead(b,3));
  digitalWrite(DATA4_PIN,bitRead(b,4));
  digitalWrite(DATA5_PIN,bitRead(b,5));
  digitalWrite(DATA6_PIN,bitRead(b,6));
  digitalWrite(DATA7_PIN,bitRead(b,7));  
}

void writeMode()
{   
   pinMode( DATA0_PIN, OUTPUT); 
   pinMode( DATA1_PIN, OUTPUT); 
   pinMode( DATA2_PIN, OUTPUT); 
   pinMode( DATA3_PIN, OUTPUT); 
   pinMode( DATA4_PIN, OUTPUT); 
   pinMode( DATA5_PIN, OUTPUT); 
   pinMode( DATA6_PIN, OUTPUT); 
   pinMode( DATA7_PIN, OUTPUT); 
}

//
//byte readByte()
//{
//  readMode(); 
//  byte result=0;
//  if( digitalRead(DATA0_PIN)) bitSet(result,0);
//  if( digitalRead(DATA1_PIN)) bitSet(result,1);
//  if( digitalRead(DATA2_PIN)) bitSet(result,2);
//  if( digitalRead(DATA3_PIN)) bitSet(result,3);
//  if( digitalRead(DATA4_PIN)) bitSet(result,4);
//  if( digitalRead(DATA5_PIN)) bitSet(result,5);
//  if( digitalRead(DATA6_PIN)) bitSet(result,6);
//  if( digitalRead(DATA7_PIN)) bitSet(result,7); 
//  return result;
//}

void readDirective(byte input)
{
  byte  directive = input & B00001111;
  if(directive ==7) acceptFrameDirective =true;  
  else acceptFrameDirective = false;

  if(directive ==6) sendFrameDirective =true;
  else sendFrameDirective = false;

  if(directive == 2 ) resetDirective =true;
  else resetDirective = false;
}

void readMode()
{    
   pinMode( DATA0_PIN, INPUT); 
   pinMode( DATA1_PIN, INPUT); 
   pinMode( DATA2_PIN, INPUT); 
   pinMode( DATA3_PIN, INPUT); 
   pinMode( DATA4_PIN, INPUT); 
   pinMode( DATA5_PIN, INPUT); 
   pinMode( DATA6_PIN, INPUT); 
   pinMode( DATA7_PIN, INPUT); 
}

void CheckSerial() 
{
    if(Serial.available() > 0) 
    {
      if( sendLength == 0)
      {     
         while (Serial.available() > 0) 
         {
          sendBytes[sendCounter] = Serial.read();
          sendLength+=1;
         }
      }      
    }
}

void UpdateStepTime()
{
  if(previousStep!=startStep) stepTime = micros();
}

void CheckStepTime(int maxTime)
{
  if(micros()>stepTime+maxTime)  ResetSequenceTimeOut();
}

//bool IA()
//{
//  return digitalRead(IA_PIN);
//}

bool WR()
{
  return digitalRead(WR_PIN);
}

bool RD()
{
  return digitalRead(RD_PIN);
}



void ProcessOutput()
{
  outputReady = true;
}
